import pandas as pd


def preprocessing(df: pd.DataFrame) -> pd.DataFrame:
    """
    Preprocessing pipeline for cluster dataframe.
    Includes handlic strings and numeric values.

    Args:
        df: pd.DataFrame - raw dataset

    Return:
        df: pd.DataFrame - preprocessed pandas dataFrame
    """
    df = df.apply(pd.to_numeric, errors="ignore")
    df = string_handling(df)
    df = numeric_handling(df)
    return df


def numeric_handling(df: pd.DataFrame) -> pd.DataFrame:
    """
    Numeric handling method. Converting string to numeric where possible

    Args:
        df - pd.DataFrame: cluster dataframe

    Return:
        df - pd.DataFrame: modified cluster dataframe
    """
    numeric_df = df.apply(pd.to_numeric, errors="coerce")
    numeric_series = numeric_df.isnull().all()
    numeric_df = numeric_df[numeric_series[~numeric_series].index]

    # Define logic separating class. Possible bug {1,2,3,"noise","target"}

    if "class" in df:
        if "class" not in numeric_df.columns:
            df = pd.concat([numeric_df, df["class"]], axis=1)
        else:
            df = numeric_df
        df["class"].fillna(-1, inplace=True)
    else:
        df = numeric_df
    return df


def string_handling(
    df: pd.DataFrame,
    decoder: str = "ascii",
    create_dummy_columns: bool = True,
    thresh_dummies: int = 5,
) -> pd.DataFrame:
    """
    String handling method. Decoding and replacing string values with one-hot encoding alternative

    Args:
        df - pd.DataFrame: cluster dataframe
        decoder - str: decoder --> ascii, utf-8, etc
        create_dummy_columns - bool: should one-hot encoding to be applied
        thresh_dummies - int: if create_dummy_columns is True, maximum number of unique values to be applied.
        Reason is to avoid creating to sparse matrix

    Return:
        df - pd.DataFrame: modified cluster dataframe
    """
    for col in df.select_dtypes("object"):

        # check for binary values
        _col = df[col].str.decode(decoder)
        if not _col.isnull().all():
            df[col] = _col

        # one-hot encoding
        if (
            create_dummy_columns
            and col != "class"
            and df[col].nunique() <= thresh_dummies
        ):
            _dummies = pd.get_dummies(df[col])
            _dummies.columns = [f"{col}_{value}" for value in _dummies.columns]
            df = pd.concat([df, _dummies], axis=1).drop([col], axis=1)
    return df


if __name__ == "__main__":
    pass



from scipy import stats
from statistics import NormalDist
import numpy as np
from . import read_utilities


class Dataset:
    def __init__(self, dataset_name, path) -> None:
        self.name = dataset_name.split(".")[-2]
        self.pd_dataframe = read_utilities.load(
            self.name, path=f"{path}/{dataset_name}"
        )
        self.description = read_utilities.load_description(
            self.name, path=f"{path}/{dataset_name}"
        )
        self.origin = path.split("/")[-1]
        self.classes = (
            self.pd_dataframe["class"].unique()
            if "class" in self.pd_dataframe.columns
            else []
        )
        self.number_of_clusters = len(self.classes)
        if self.number_of_clusters == 1 and "class" in self.pd_dataframe.columns:
            self.pd_dataframe.drop("class", inplace=True, axis=1)
        self.dimensions = [x for x in self.pd_dataframe if x != "class"]
        self.std_dictionary, self.mean_dictionary = self.calculate_std_mean()
        self.number_of_datapoint = self.calculate_dataframe_shape(cols=False)
        self.number_of_dimensions = self.calculate_dataframe_shape(rows=False) - int(
            "class" in self.pd_dataframe.columns
        )
        self.missing_values_proportion = np.round(self.calculate_number_of_nan(), 4)
        self.is_imbalanced = None
        self.estimate_max_overlap, self.estimate_mean_overlap = None, None
        if self.number_of_clusters > 1:
            self.is_imbalanced = bool(self.calculate_is_imbalanced())
            (
                self.estimate_max_overlap,
                self.estimate_mean_overlap,
            ) = self.calculate_estimate_overlap()
        self.strong_outlier_proportion = np.round(
            self.calculate_number_of_outliers(number_of_std=3), 4
        )
        self.weak_outlier_proportion = np.round(
            (
                self.calculate_number_of_outliers(number_of_std=2)
                - self.strong_outlier_proportion
            ),
            4,
        )

    def calculate_number_of_nan(self):
        return (
            self.pd_dataframe.isnull().sum().sum()
            / self.pd_dataframe.shape[0]
            / self.pd_dataframe.shape[1]
        )

    def calculate_std_mean(self):
        if self.number_of_clusters > 1:
            return self.pd_dataframe.groupby("class").std().to_dict(
                "index"
            ), self.pd_dataframe.groupby("class").mean().to_dict("index")
        else:
            return self.pd_dataframe.std().to_dict(), self.pd_dataframe.mean().to_dict()

    def calculate_estimate_overlap(self):
        overlap_list = []
        for idx, cluster1 in enumerate(self.classes):
            for cluster2 in self.classes[idx + 1 :]:
                for dimension in self.dimensions:
                    overlap = NormalDist(
                        mu=self.mean_dictionary[cluster1][dimension],
                        sigma=np.nan_to_num(self.std_dictionary[cluster1][dimension])
                        + 1e-10,
                    ).overlap(
                        NormalDist(
                            mu=self.mean_dictionary[cluster2][dimension],
                            sigma=np.nan_to_num(
                                self.std_dictionary[cluster2][dimension]
                            )
                            + 1e-10,
                        )
                    )
                overlap_list.append(np.prod(overlap))
        return np.round(np.max(overlap_list), 4), np.round(np.mean(overlap_list), 4)

    def calculate_dataframe_shape(self, rows=True, cols=True):
        if rows:
            if cols:
                return self.pd_dataframe.shape
            return self.pd_dataframe.shape[0]
        elif cols:
            return self.pd_dataframe.shape[1]
        return None

    def calculate_is_imbalanced(
        self, shapiro_alpha: float = 0.05, binary_alpha=0.2
    ) -> bool:
        if self.number_of_clusters == 2:
            return (
                self.pd_dataframe["class"].value_counts().min()
                / self.number_of_datapoint
                > binary_alpha
            )
        class_frequency = self.pd_dataframe["class"].value_counts().to_list()
        if [class_frequency[0]] * self.number_of_clusters == class_frequency:
            return False
        else:
            _, p = stats.shapiro(class_frequency)
            return p < shapiro_alpha

    def calculate_number_of_outliers(self, number_of_std=2):
        if self.number_of_clusters > 1:
            return (
                self.pd_dataframe.groupby("class")
                .apply(lambda x: np.abs(stats.zscore(x) > number_of_std))
                .any(axis=1)
                .sum()
                / self.number_of_datapoint
            )
        else:
            return (
                self.pd_dataframe.apply(lambda x: np.abs(stats.zscore(x) > 2))
                .any(axis=1)
                .sum()
                / self.number_of_datapoint
            )


if __name__ == "__main__":
    pass



#!/usr/bin/env python3

from typing import List
import pandas as pd
import numpy as np
from .utilities import read_utilities, anomaly_detection
from sklearn.cluster import (
    KMeans,
    DBSCAN,
    AgglomerativeClustering,
    Birch,
    AffinityPropagation,
    FeatureAgglomeration,
    MiniBatchKMeans,
    SpectralClustering,
    MeanShift,
    BisectingKMeans,
    OPTICS,
    SpectralBiclustering,
    SpectralCoclustering,
)
from sklearn.mixture import GaussianMixture


class Dataset:
    def __init__(self) -> None:
        pass

    def load_stats(self, dataset_name: str) -> dict or None:
        """
        Load specific dataset statistics from local.
        Statistics are calculated using the dataset_utilities

        Args:
            dataset_name: str - name of dataset

        Return:
            dict or None - dataset_info.json content or None if dataset is missing in catalogue
        """
        dataset_info = read_utilities.load_dataset_info()
        if dataset_name not in dataset_info:
            print(
                f"Dataset {dataset_name} not present in the local catalogue. To update catalogue, run update_local_info()"
            )
            return None
        else:
            return dataset_info[dataset_name]["stats"]

    def load_description(self, dataset_name: str):
        """
        Load specific dataset descritption from local.
        Description is given by authors and contributors of the dataset

        Args:
            dataset_name: str - name of dataset

        Return:
            dict or None - description provided or None if dataset is missing in catalogue
        """
        dataset_info = read_utilities.load_dataset_info()
        if dataset_name not in dataset_info:
            print(
                f"Dataset {dataset_name} not present in the local catalogue. To update catalogue, run update_local_info()"
            )
            return None
        else:
            return dataset_info[dataset_name]["description"]

    def load(
        self, dataset_name: str, download=False, overwrite=False
    ) -> pd.DataFrame or None:
        """
        Load specific dataset from local. Optionally one can download dataset if specified to local

        Args:
            dataset_name: str - name of dataset
            download: bool - download dataset if missing locally on demand
            overwrite: bool - overwrite downloaded dataset. Used only if download==True

        Return:
            pd.DataFrame or None - dataset data in pd.DataFrame or None if dataset is missing locally or in catalogue
        """
        dataset_info = read_utilities.load_dataset_info()
        if dataset_name not in dataset_info:
            print(
                f"Dataset {dataset_name} not present in the local catalogue. To update catalogue, run update_local_info()"
            )
            return None
        if dataset_name not in self.list(is_print=False):
            if download:
                self.download(dataset_name, overwrite)
            else:
                print(
                    f"Dataset {dataset_name} isn't present in local environment. To allow downloading, specify load(download=True)"
                )
                return None
        return read_utilities.load(dataset_name)

    def download(self, dataset_names: str or List, overwrite=False) -> None:
        """
        Download dataset from source repo. One can pass single dataset name or list of names.

        Args:
            datasets: str or List[str] - one or single dataset names
            overwrite: bool - overwrite if dataset is present
        """
        read_utilities.download(dataset_names, overwrite=overwrite)

    def list(self, is_print: bool = False) -> List:
        """
        Listing all available datasets locally.

        Args:
            is_print: bool - print datasets list if True

        Return:
            filename: List - list of all locally avaliable datasets
        """
        datasets = read_utilities.list_local_datasets()
        if is_print:
            print(datasets)
        return datasets

    def update_local_info_files(self):
        """
        Update local files from github source
        """
        read_utilities.update_local_jsons()


class Methods:
    def __init__(self) -> None:
        pass

    def KMeans(*args):
        return KMeans(args)

    def MiniBatchKMeans(*args):
        return MiniBatchKMeans(args)

    def BisectingKMeans(*args):
        return BisectingKMeans(args)

    def DBSCAN(*args):
        return DBSCAN(args)

    def AgglomerativeClustering(*args):
        return AgglomerativeClustering(args)

    def GaussianMixture(*args):
        return GaussianMixture(args)

    def Birch(*args):
        return Birch(args)

    def AffinityPropagation(*args):
        return AffinityPropagation(args)

    def FeatureAgglomeration(*args):
        return FeatureAgglomeration(args)

    def OPTICS(*args):
        return OPTICS(args)

    def MeanShift(*args):
        return MeanShift(args)

    def SpectralClustering(*args):
        return SpectralClustering(args)

    def SpectralBiclustering(*args):
        return SpectralBiclustering(args)

    def SpectralCoclustering(*args):
        return SpectralCoclustering(args)


class Tasks:
    def __init__(self) -> None:
        pass

    def detect_anomalies(
        self,
        dataset: np.array,
        methods: str = "all_besides_nn",
        mode: str = "per_class",
        outliers_fraction: float = 0.1,
    ):
        """
        Detect anomalies in the passed dataset
        """
        return anomaly_detection.detect_anomalies(
            dataset, methods, mode, outliers_fraction
        )

    def plot_overall_anomaly_classifiers(self, result, classifiers, show_heatmap=True):
        return anomaly_detection.plot_overall(
            result, classifiers, show_heatmap=show_heatmap
        )

    def plot_individual_anomaly_classifiers(
        self, result, classifiers, detected_outliers
    ):
        return anomaly_detection.plot_classifiers(
            result, classifiers, detected_outliers
        )

    def suggest_anomaly_detection_method(self, dataset):
        pass

    def suggest_clustering_method(self, dataset):
        pass

    def calculate_number_of_clusters(self, dataset, methods=["kmeans"]):
        """
        Calculater the number of clusters in the passed dataset
        """
        pass

    def cluster_ensembling(self, dataset, methods=["kmeans", "gmm", "dbscan"]):
        """Provide cluster ensempling"""
        pass

    def cluster_similarity(self, dataset, method="knn", nearest_neighbors=3):
        """Provides similar known datasets to provided one using knn"""
        pass

    def classify_semi_labeled_data(
        self,
    ):
        pass

    def one_class_clustering(
        self,
    ):
        pass



from typing import List
import arff
import os
from scipy.io import arff as scipy_arff
import json
import pandas as pd
import requests
from . import preprocessing_utilities


def read_arff(filepath: str, mode: str = "scipy_arff"):
    """
    Reading aiff format.
    By default scipy.io arff reader is used (mode: scipy_arff).

    Args:
        filepath - str: full path to aiff file
        mode - str: scipy_arff | arff

    Return:
        arff reade format
    """
    if mode == "scipy_arff":
        return scipy_arff.loadarff(filepath)
    elif mode == "arff":
        return arff.load(open(filepath, "r"))
    return None


def get_root_folder() -> str:
    """
    Get root folder

    Return:
        os.path: root folder
    """
    return os.path.dirname(os.path.dirname(__file__))


def get_dataset_info(name: str):
    """
    Get information regarding particular {name} dataset

    Args:
        name: str

    Return:
        dict or None: specified dict by name from dataset_info.json
    """
    data = load_dataset_info()
    if name in data:
        return data[name]
    else:
        print(
            f"Dataset {name} not present in the local catalogue. To update catalogue, run update_local_info()"
        )
        return None


def load_catalogue_info(root_folder: str = None) -> dict:
    """
    Loading content of catalogue_info.json.
    Contains information regarding datasets paths and configs

    Args:
        root_folder:str - root folder

    Return:
        dict: catalogue_info.json content
    """
    if not root_folder:
        root_folder = get_root_folder()
    return json.load(open(f"{root_folder}/jsons/catalogue_info.json"))


def load_dataset_info(root_folder: str = None) -> dict:
    """
    Loading content of dataset_info.json.
    Contains information regarding datasets paths and configs

    Args:
        root_folder:str - root folder

    Return:
        dict: dataset_info.json content
    """
    if not root_folder:
        root_folder = get_root_folder()
    return json.load(open(f"{root_folder}/jsons/dataset_info.json"))


def load_description(name: str, path: str = "") -> str or None:
    """
    Loading dataset decsription by name

    Args:
        name : str - dataset name

    Return:
        description: str or None - cluster dataset in pandas Dataframe.
        None if file isn't found in dataset folder.

    Errors:
        NotImplementedError - if some functionality isn't implemented in scipy.io arff method
        FileNotFoundError - dataset file isn't present
    """
    if not path:
        root_folder = get_root_folder()
        dataset_info = get_dataset_info(name)
        path = f'{root_folder}/{dataset_info["local_filepath"]}'
    try:
        data = read_arff(f"{path}")
        description = pd.DataFrame(data[1])
    except NotImplementedError:
        data = read_arff(f"{path}", mode="arff")
        description = pd.DataFrame(data["description"])
    except FileNotFoundError:
        print(
            f"Dataset {name} not present in the local catalogue. To update catalogue, run update_local_info()"
        )
        return None
    return description


def load(name: str, path: str = "") -> pd.DataFrame or None:
    """
    Loading dataset by name

    Args:
        name : str - dataset name

    Return:
        df: pd.DataFrame or None - cluster dataset in pandas Dataframe.
        None if file isn't found in dataset folder.

    Errors:
        NotImplementedError - if some functionality isn't implemented in scipy.io arff method
        FileNotFoundError - dataset file isn't present
    """
    if not path:
        root_folder = get_root_folder()
        dataset_info = get_dataset_info(name)
        path = f'{root_folder}/{dataset_info["local_filepath"]}'
    try:
        data = read_arff(f"{path}")
        df = pd.DataFrame(data[0])
        df.columns = df.columns.str.lower()
    except NotImplementedError:
        data = read_arff(f"{path}", mode="arff")
        df = pd.DataFrame(
            data["data"], columns=[x[0].lower() for x in data["attributes"]]
        )
    except FileNotFoundError:
        print(
            f"Dataset {name} not present in the local catalogue. To update catalogue, run update_local_info()"
        )
        return None
    df = preprocessing_utilities.preprocessing(df)
    return df


def download(datasets: str or List[str], overwrite: bool = False):
    """
    Download dataset from github repo.

    Args:
        datasets: str or List[str] - one or single dataset names
        overwrite: bool - overwrite if dataset is present
    """
    if isinstance(datasets, str):
        datasets = [datasets]
    status_datasets = {
        "Dataset_not_found_in_catalogue": [],
        "Download_success": [],
        "Filepath_not_valid": [],
    }
    root_folder = get_root_folder()
    dataset_info = load_dataset_info(root_folder)
    for dataset in datasets:
        if dataset not in dataset_info:
            status_datasets["Dataset_not_found_in_catalogue"].append(dataset)
        else:
            github_path = dataset_info[dataset]["github_filepath"]
            r = requests.get(github_path, allow_redirects=True)
            if r.status_code != 200:
                status_datasets["Filepath_not_valid"].append(dataset)
            open(
                f'{root_folder}/datasets/{dataset_info[dataset]["name"]}.{dataset_info[dataset]["filetype"]}',
                "w",
            ).write(r.text)
            status_datasets["Download_success"].append(dataset)

    for status in status_datasets:
        if status_datasets[status]:
            print(f"{status}: {status_datasets[status]}")


def list_local_datasets() -> List:
    """
    Listing all available datasets locally.

    Return:
        filename: List - list of all locally avaliable datasets
    """
    root_folder = get_root_folder()
    catalogue_info = load_catalogue_info(root_folder)
    dataset_info = load_dataset_info(root_folder)
    local_filepath_dict = {
        dataset_info[filename]["local_filepath"]: filename for filename in dataset_info
    }
    filenames = [
        local_filepath_dict[f'{catalogue_info["PATH_TO_LOCAL"]}/{x}']
        if f'{catalogue_info["PATH_TO_LOCAL"]}/{x}' in local_filepath_dict
        else x
        for x in os.listdir(f'{root_folder}/{catalogue_info["PATH_TO_LOCAL"]}')
    ]
    return filenames


def update_local_jsons():
    root_folder = get_root_folder()
    catalogue_info = load_catalogue_info()
    github_path = catalogue_info["PATH_TO_GITHUB"]
    for file in ["catalogue_info.json", "dataset_info.json"]:
        r = requests.get(f"{github_path}/{file}", allow_redirects=True)
        if r.status_code == 200:
            with open(f"{root_folder}/jsons/{file}", "w") as outfile:
                json.dump(r.json(), outfile)
            print(f"{file} successfully updated")
        else:
            print(f"{file} wasn't updated")


if __name__ == "__main__":
    pass



from asyncio import tasks
import python_clustering


def main():
    print(dir(python_clustering))
    dataset = python_clustering.Dataset()
    dataset_list = ["atom", "D31", "cpu", "banana", "gaussians1", "circle"]
    # assert(dataset.list() == dataset_list)
    data = dataset.load("D31")
    print(data.shape)
    data = dataset.load("D3aa1")
    print(data)
    data = dataset.load("blobs")
    print(data)
    data = dataset.load("blobs", download=True)
    print(data)
    dataset.download("rings")
    description = dataset.load_description("blobs")
    print(description)
    stats = dataset.load_stats("rings")
    print(stats)
    dataset.update_local_info_files()

    tasks = python_clustering.Tasks()
    data = dataset.load("D31").values
    anomalies, methods_results = tasks.detect_anomalies(data)

main()



from typing import List
import matplotlib
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
import numpy as np

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

from pyod.models.ecod import ECOD
from pyod.models.copod import COPOD
from pyod.models.abod import ABOD
from pyod.models.sos import SOS
from pyod.models.gmm import GMM
from pyod.models.sampling import Sampling
from pyod.models.kde import KDE

from pyod.models.ocsvm import OCSVM
from pyod.models.pca import PCA
from pyod.models.mcd import MCD

from pyod.models.cblof import CBLOF
from pyod.models.hbos import HBOS
from pyod.models.knn import KNN
from pyod.models.lof import LOF

from pyod.models.loda import LODA
from pyod.models.inne import INNE
from pyod.models.rod import ROD
from pyod.models.iforest import IForest

from pyod.models.auto_encoder import AutoEncoder
from pyod.models.so_gaal import SO_GAAL
from pyod.models.mo_gaal import MO_GAAL
from pyod.models.deep_svdd import DeepSVDD
from pyod.models.anogan import AnoGAN


def build_classifiers(
    methods: str or List = "all",
    classifiers={},
    outliers_fraction=0.1,
    random_state=42,
    n_neighbors=25,
    n_estimators=250,
):
    all_methods = True if methods == "all" else False
    all_besides_nn_methods = True if methods == "all_besides_nn" else False

    # Probabilistic
    if all_methods or all_besides_nn_methods or "ECOD" in methods:
        classifiers["Empirical Cumulative Distribution Functions (ECOD)"] = ECOD(
            contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "COPOD" in methods:
        classifiers["Copula Based Outlier Detector (COPOD)"] = COPOD(
            contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "ABOD" in methods:
        classifiers["Angle-based Outlier Detector (ABOD)"] = ABOD(
            contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "ABOD" in methods:
        classifiers["Stochastic Outlier Selection (SOS)"] = SOS(
            contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "KDE" in methods:
        classifiers["Kernel Density Estimation (KDE)"] = KDE(
            contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "GMM" in methods:
        classifiers["Gaussian Mixture Models (GMM)"] = GMM(
            contamination=outliers_fraction, random_state=random_state
        )
    if all_methods or all_besides_nn_methods or "SP" in methods:
        classifiers["Outlier detection based on Sampling (SP)"] = Sampling(
            contamination=outliers_fraction
        )

    # Linear
    if all_methods or all_besides_nn_methods or "PCA" in methods:
        classifiers["Principal Component Analysis (PCA)"] = PCA(
            contamination=outliers_fraction, random_state=random_state
        )
    if all_methods or all_besides_nn_methods or "OCSVM" in methods:
        classifiers["One-class SVM (OCSVM)"] = OCSVM(contamination=outliers_fraction)
    if all_methods or all_besides_nn_methods or "MCD" in methods:
        classifiers["Minimum Covariance Determinant (MCD)"] = MCD(
            contamination=outliers_fraction, random_state=random_state
        )

    # Proximity-Based
    if all_methods or all_besides_nn_methods or "CBLOF" in methods:
        classifiers["Cluster-based Local Outlier Factor (CBLOF)"] = CBLOF(
            contamination=outliers_fraction,
            check_estimator=False,
            random_state=random_state,
        )
    if all_methods or all_besides_nn_methods or "LOF" in methods:
        classifiers["Local Outlier Factor (LOF)"] = LOF(
            n_neighbors=n_neighbors, contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "HBOS" in methods:
        classifiers["Histogram-base Outlier Detection (HBOS)"] = HBOS(
            contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "KNN" in methods:
        classifiers["K Nearest Neighbors (KNN)"] = KNN(
            method="largest", contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "mKNN" in methods:
        classifiers["Mean K Nearest Neighbors (mKNN)"] = KNN(
            method="mean", contamination=outliers_fraction
        )
    if all_methods or all_besides_nn_methods or "ROD" in methods:
        classifiers["Rotation-based Outlier Detector (ROD)"] = ROD(
            contamination=outliers_fraction
        )

    # Outlier Ensembles
    if all_methods or all_besides_nn_methods or "IF" in methods:
        classifiers["Isolation Forest (IF)"] = IForest(
            n_estimators=n_estimators,
            contamination=outliers_fraction,
            random_state=random_state,
        )
    if all_methods or all_besides_nn_methods or "INNE" in methods:
        classifiers[
            "Isolation-based anomaly detection using nearest-neighbor ensembles (INNE)"
        ] = INNE(
            contamination=outliers_fraction,
            n_estimators=n_estimators,
            random_state=random_state,
        )
    if all_methods or all_besides_nn_methods or "LODA" in methods:
        classifiers["Lightweight on-line detector of anomalies (LODA)"] = LODA(
            contamination=outliers_fraction
        )

    # Neural Networks

    if all_methods or "AutoEncoder" in methods:
        classifiers["AutoEncoder"] = AutoEncoder(
            contamination=outliers_fraction,
            random_state=random_state,
            hidden_neurons=[0],
            verbose=0,
        )
    if all_methods or "SOGAAL" in methods:
        classifiers[
            "Single-Objective Generative Adversarial Active Learning (SOGAAL)"
        ] = SO_GAAL(contamination=outliers_fraction)
    if all_methods or "MOGAAL" in methods:
        classifiers[
            "Multiple-Objective Generative Adversarial Active Learning (MOGAAL)"
        ] = MO_GAAL(contamination=outliers_fraction)
    if all_methods or "DeepSVDD" in methods:
        classifiers["Deep One-Class Classifier with AutoEncoder (DeepSVDD)"] = DeepSVDD(
            contamination=outliers_fraction, verbose=0, random_state=random_state
        )
    if all_methods or "AnoGAN" in methods:
        classifiers["Anomaly Detection with Generative Adversarial Networks"] = AnoGAN(
            contamination=outliers_fraction, verbose=0
        )
    return classifiers


def predict_outliers(clf, X, raw=False):
    clf.fit(X)
    if raw:
        return clf.decision_function(X)
    return clf.predict(X)


def get_number_of_methods(classifiers):
    return len(classifiers)


def detect_outliers(X, classifiers):
    results = {clf: predict_outliers(classifiers[clf], X) for clf in classifiers}
    return results


def aggregate_outliers(results):
    return np.array(list(results.values())).sum(axis=0) / len(results)


def run_ontlier_detection(X, classifiers, mode="per_class"):
    final_arr = np.array([])
    if mode == "per_class":
        detected_outliers_dict = {}
        for _class in np.unique(X[:, 2]):
            arr = X[X[:, 2] == _class]
            detected_outliers_dict[_class] = detect_outliers(arr[:, 0:2], classifiers)
            overall = aggregate_outliers(detected_outliers_dict[_class])
            if final_arr.size:
                final_arr = np.vstack(
                    (final_arr, np.hstack((arr, overall.reshape(-1, 1))))
                )
            else:
                final_arr = np.hstack((arr, overall.reshape(-1, 1)))
    else:
        detected_outliers_dict = detect_outliers(X[:, 0:2], classifiers)
        overall = aggregate_outliers(detected_outliers_dict)
        final_arr = np.hstack((X, overall.reshape(-1, 1)))
    return final_arr, detected_outliers_dict


def plot_classifiers(
    X, classifiers, detected_outliers=None, increase_coef=0.05, plot_per_row=3
):
    """
    X[:,:-2] - data columns
    X[:,-2]  - class column
    X[:,-1]  - proba column
    """
    min_x, max_x = X[:, 0].min() - abs(increase_coef * X[:, 0].min()), X[
        :, 0
    ].max() + abs(increase_coef * X[:, 0].max())
    min_y, max_y = X[:, 1].min() - abs(increase_coef * X[:, 1].min()), X[
        :, 1
    ].max() + abs(increase_coef * X[:, 1].max())
    xx, yy = np.meshgrid(np.linspace(min_x, max_x, 100), np.linspace(min_y, max_y, 100))

    figsize = (15, int(len(classifiers) / plot_per_row) * 3)
    plt.figure(figsize=figsize)

    for i, (clf_name, clf) in enumerate(classifiers.items()):
        y_pred = detected_outliers[clf_name]
        X_inline = X[np.where(y_pred == 0)][:, :-2]
        X_outline = X[np.where(y_pred == 1)][:, :-2]
        scores_pred = clf.decision_function(X[:, :-2]) * -1

        Z = clf.decision_function(np.c_[xx.ravel(), yy.ravel()]) * -1
        Z = Z.reshape(xx.shape)
        scores_pred = np.maximum(scores_pred, Z)

        threshold = np.percentile(scores_pred, X_outline.shape[0] * 100 / X.shape[0])

        subplot = plt.subplot(
            int(len(classifiers) / plot_per_row) + 1, plot_per_row, i + 1
        )
        subplot.contourf(
            xx, yy, Z, levels=np.linspace(Z.min(), threshold, 7), cmap=plt.cm.YlGnBu_r
        )
        a = subplot.contour(xx, yy, Z, levels=[threshold], linewidths=2, colors="green")
        subplot.contourf(xx, yy, Z, levels=[threshold, Z.max()], colors="orange")

        b = subplot.scatter(
            X_inline[:, 0], X_inline[:, 1], c="white", s=50, edgecolor="k"
        )
        c = subplot.scatter(
            X_outline[:, 0], X_outline[:, 1], c="red", s=50, edgecolor="k"
        )
        subplot.axis("tight")
        subplot.legend(
            [b, c],
            ["pred inliers", "pred outliers"],
            prop=matplotlib.font_manager.FontProperties(size=10),
            loc="lower right",
        )
        subplot.set_xlabel(f"{i}, {clf_name}")
        subplot.set_xlim((min_x, max_x))
        subplot.set_ylim((min_y, max_y))
    plt.subplots_adjust(0.1, 0.1, 1, 1, 0.1, 0.4)


def plot_overall(
    X,
    classifiers,
    calculated_class=None,
    show_heatmap=False,
    increase_coef=0.05,
    figsize=(12, 8),
):
    plt.figure(figsize=figsize)
    if not calculated_class:
        calculated_class = X[:, -1]
    min_x, max_x = (
        X[:, 0].min() - increase_coef * X[:, 0].min(),
        X[:, 0].max() + increase_coef * X[:, 0].max(),
    )
    min_y, max_y = (
        X[:, 1].min() - increase_coef * X[:, 1].min(),
        X[:, 1].max() + increase_coef * X[:, 1].max(),
    )

    if show_heatmap:
        k = gaussian_kde([X[:, 0], X[:, 1]])
        xi, yi = np.mgrid[
            min_x : max_x : X[:, 0].size * 1j * 3, min_y : max_y : X[:, 1].size * 1j * 3
        ]
        zi = k(np.vstack([xi.flatten(), yi.flatten()]))
        plt.pcolormesh(xi, yi, zi.reshape(xi.shape), shading="auto", cmap="YlGnBu_r")

    plt.scatter(X[:, 0], X[:, 1], s=75, c=calculated_class, cmap="Reds", edgecolor="k")
    cbar = plt.colorbar()
    cbar.set_label("Outlier probability", rotation=270, labelpad=15)
    plt.xlim((min_x, max_x))
    plt.ylim((min_y, max_y))
    plt.title(
        f"Overall outlier map based on {get_number_of_methods(classifiers)} methods"
    )


def detect_anomalies(
    dataset: np.array,
    methods: str = "all_besides_nn",
    mode: str = "per_class",
    outliers_fraction=0.05,
):
    classifiers = build_classifiers(
        methods=methods, outliers_fraction=outliers_fraction
    )
    result, detected_outliers = run_ontlier_detection(dataset, classifiers, mode=mode)
    return result, detected_outliers



# python3 utilities/catalogue_utilities.py

import os
import json
from pytablewriter import MarkdownTableWriter
from . import dataset_utilities


class Catalogue:
    def __init__(self):
        data = json.load(open("./jsons/catalogue_info.json"))
        self.PATH_TO_ALL_DATASETS_FOLDER = data["PATH_TO_ALL_DATASETS_FOLDER"]
        self.PATH_TO_LOCAL = data["PATH_TO_LOCAL"]
        self.PATH_TO_GITHUB = data["PATH_TO_GITHUB"]
        self.SUBFOLDERS = data["SUBFOLDERS"]
        self.COLUMN_NAMES = data["COLUMN_PROPERTY_NAMES"]
        self.ROOT_NAMES = data["ROOT_PROPERTY_NAMES"]

    def append(self, filename, subfolder):
        path = f"{self.PATH_TO_ALL_DATASETS_FOLDER}/{subfolder}"
        dataset = dataset_utilities.Dataset(filename, path=path)
        with open(f"{self.PATH_TO_ALL_DATASETS_FOLDER}/dataset_info.json", "r") as f:
            catalogue = json.load(f)
        catalogue[filename.split(".")[-2]] = self.arrange_calculated_data(
            dataset, subfolder=subfolder, filename=filename
        )
        self.create_markdown_table(catalogue)
        self.create_dataset_info(catalogue)

    def create(self):
        catalogue = {}
        for subfolder in self.SUBFOLDERS:
            path = f"{self.PATH_TO_ALL_DATASETS_FOLDER}/{subfolder}"
            for filename in os.listdir(path):
                dataset = dataset_utilities.Dataset(filename, path=path)
                catalogue[filename.split(".")[-2]] = self.arrange_calculated_data(
                    dataset, subfolder=subfolder, filename=filename
                )
        self.create_markdown_table(catalogue)
        self.create_dataset_info(catalogue)

    def create_dataset_info(self, catalogue):
        with open(
            f"{self.PATH_TO_ALL_DATASETS_FOLDER}/dataset_info.json", "w"
        ) as json_file:
            json.dump(catalogue, json_file)

    def create_markdown_table(self, data):
        value_matrix = [[]] * len(data)
        for idx_file, file in enumerate(data):
            row = [None] * len(self.COLUMN_NAMES)
            for idx_col, col in enumerate(self.COLUMN_NAMES):
                if col in data[file]:
                    row[idx_col] = data[file][col]
                else:
                    row[idx_col] = data[file]["stats"][col]
            value_matrix[idx_file] = row
        writer = MarkdownTableWriter(
            table_name="Dataset Description",
            headers=self.COLUMN_NAMES,
            value_matrix=value_matrix,
        )
        with open(f"{self.PATH_TO_ALL_DATASETS_FOLDER}/README.md", "w") as f:
            writer.stream = f
            writer.write_table()
            f.write("\n")
            f.write("* anomaly percentage were calculated using zscore")

    def arrange_calculated_data(self, dataset, subfolder, filename):
        dataset_dict = {"stats": {}}
        for key, value in self.collect_calculated_data(
            dataset, subfolder, filename
        ).items():
            if key in self.ROOT_NAMES:
                dataset_dict[key] = value
            else:
                dataset_dict["stats"][key] = value
        return dataset_dict

    def collect_calculated_data(self, dataset, subfolder, filename):
        return {
            "name": dataset.name,
            "origin": dataset.origin,
            "filetype": filename.split(".")[-1],
            "local_filepath": f"{self.PATH_TO_LOCAL}/{filename}",
            "github_filepath": f"{self.PATH_TO_GITHUB}/{subfolder}/{filename}",
            "description": str(dataset.description),
            "is_imbalanced": dataset.is_imbalanced,
            "number_of_clusters": dataset.number_of_clusters,
            "number_of_datapoint": dataset.number_of_datapoint,
            "number_of_dimensions": dataset.number_of_dimensions,
            "weak_outlier_proportion": dataset.weak_outlier_proportion,
            "strong_outlier_proportion": dataset.strong_outlier_proportion,
            "estimated_max_overlap": dataset.estimate_max_overlap,
            "estimated_mean_overlap": dataset.estimate_mean_overlap,
            "missing_values_proportion": dataset.missing_values_proportion,
        }


if __name__ == "__main__":
    catalogue = Catalogue()
    catalogue.create()
